@@ -22,15 +22,14 @@ import (
 
 	"github.com/buildkite/agent/v3/agent/plugin"
 	"github.com/buildkite/agent/v3/env"
-	"github.com/buildkite/agent/v3/internal/experiments"
 	"github.com/buildkite/agent/v3/internal/file"
 	"github.com/buildkite/agent/v3/internal/job/hook"
-	"github.com/buildkite/agent/v3/internal/job/shell"
+	"github.com/buildkite/agent/v3/internal/osutil"
 	"github.com/buildkite/agent/v3/internal/redact"
 	"github.com/buildkite/agent/v3/internal/replacer"
+	"github.com/buildkite/agent/v3/internal/shell"
 	"github.com/buildkite/agent/v3/internal/shellscript"
 	"github.com/buildkite/agent/v3/internal/tempfile"
-	"github.com/buildkite/agent/v3/internal/utils"
 	"github.com/buildkite/agent/v3/kubernetes"
 	"github.com/buildkite/agent/v3/process"
 	"github.com/buildkite/agent/v3/tracetools"
@@ -84,29 +83,33 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 	ctx, cancel := context.WithCancel(ctx)
 	defer cancel()
 
-	// Check if not nil to allow for tests to overwrite shell
-	if e.shell == nil {
-		var err error
-		logger := shell.StderrLogger
-		logger.DisabledWarningIDs = e.DisabledWarnings
-		e.shell, err = shell.New(shell.WithLogger(logger))
-		if err != nil {
-			fmt.Printf("Error creating shell: %v", err)
-			return 1
-		}
+	// Start with stdout and stderr as their usual selves.
+	stdout, stderr := io.Writer(os.Stdout), io.Writer(os.Stderr)
 
-		e.shell.PTY = e.ExecutorConfig.RunInPty
-		e.shell.Debug = e.ExecutorConfig.Debug
-		e.shell.InterruptSignal = e.ExecutorConfig.CancelSignal
-		e.shell.SignalGracePeriod = e.ExecutorConfig.SignalGracePeriod
-	}
+	// The shell environment is initially the current environment.
+	// It is mutated by kubernetesSetup and needed for setupRedactors.
+	environ := env.FromSlice(os.Environ())
+
+	// Create a logger to stderr that can be used for things prior to the
+	// redactor setup.
+	// Be careful not to log customer secrets here!
+	tempLog := shell.NewWriterLogger(stderr, true, e.DisabledWarnings)
 
 	if e.KubernetesExec {
+		tempLog.Commentf("Using Kubernetes support")
+
 		socket := &kubernetes.Client{ID: e.KubernetesContainerID}
-		if err := e.kubernetesSetup(ctx, socket); err != nil {
-			e.shell.Errorf("Failed to start kubernetes socket client: %v", err)
+		if err := e.kubernetesSetup(ctx, environ, socket); err != nil {
+			tempLog.Errorf("Failed to start kubernetes socket client: %v", err)
 			return 1
 		}
+
+		// Tee both stdout and stderr to the k8s socket client, so that the
+		// logs are shipped to the agent container and then to Buildkite, but
+		// are also visible as container logs.
+		stdout = io.MultiWriter(stdout, socket)
+		stderr = io.MultiWriter(stderr, socket)
+
 		defer func() {
 			_ = socket.Exit(exitCode)
 		}()
@@ -114,7 +117,26 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 
 	// setup the redactors here once and for the life of the executor
 	// they will be flushed at the end of each hook
-	e.setupRedactors()
+	preRedactedStdout, preRedactedLogger := e.setupRedactors(tempLog, environ, stdout, stderr)
+
+	// Check if not nil to allow for tests to overwrite shell.
+	if e.shell == nil {
+		sh, err := shell.New(
+			shell.WithDebug(e.ExecutorConfig.Debug),
+			shell.WithEnv(environ),
+			shell.WithLogger(preRedactedLogger), // shell -> logger -> redactor -> real stderr
+			shell.WithInterruptSignal(e.ExecutorConfig.CancelSignal),
+			shell.WithPTY(e.ExecutorConfig.RunInPty),
+			shell.WithStdout(preRedactedStdout), // shell -> redactor -> real stdout
+			shell.WithSignalGracePeriod(e.ExecutorConfig.SignalGracePeriod),
+			shell.WithTraceContextCodec(e.TraceContextCodec),
+		)
+		if err != nil {
+			fmt.Printf("Error creating shell: %v", err)
+			return 1
+		}
+		e.shell = sh
+	}
 
 	var err error
 	span, ctx, stopper := e.startTracing(ctx)
@@ -154,7 +176,7 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 			e.shell.Errorf("Error tearing down job executor: %v", err)
 
 			// this gets passed back via the named return
-			exitCode = shell.GetExitCode(err)
+			exitCode = shell.ExitCode(err)
 		}
 	}()
 
@@ -168,21 +190,21 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 		err := e.configureGitCredentialHelper(ctx)
 		if err != nil {
 			e.shell.Errorf("Error configuring git credential helper: %v", err)
-			return shell.GetExitCode(err)
+			return shell.ExitCode(err)
 		}
 
 		// so that the new credential helper will be used for all github urls
 		err = e.configureHTTPSInsteadOfSSH(ctx)
 		if err != nil {
 			e.shell.Errorf("Error configuring https instead of ssh: %v", err)
-			return shell.GetExitCode(err)
+			return shell.ExitCode(err)
 		}
 	}
 
 	// Initialize the environment, a failure here will still call the tearDown
 	if err = e.setUp(ctx); err != nil {
 		e.shell.Errorf("Error setting up job executor: %v", err)
-		return shell.GetExitCode(err)
+		return shell.ExitCode(err)
 	}
 
 	// Execute the job phases in order
@@ -249,7 +271,7 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 				// error reporting below.
 			} else {
 				// Only upload has errored, report its error.
-				return shell.GetExitCode(err)
+				return shell.ExitCode(err)
 			}
 		}
 	}
@@ -259,7 +281,7 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 	if phaseErr != nil {
 		err = phaseErr
 		e.shell.Errorf("%v", phaseErr)
-		return shell.GetExitCode(phaseErr)
+		return shell.ExitCode(phaseErr)
 	}
 
 	// Use the exit code from the command phase
@@ -338,7 +360,7 @@ func (e *Executor) executeHook(ctx context.Context, hookCfg HookConfig) error {
 	}
 	hookName += " " + hookCfg.Name
 
-	if !utils.FileExists(hookCfg.Path) {
+	if !osutil.FileExists(hookCfg.Path) {
 		if e.Debug {
 			e.shell.Commentf("Skipping %s hook, no script at \"%s\"", hookName, hookCfg.Path)
 		}
@@ -347,10 +369,6 @@ func (e *Executor) executeHook(ctx context.Context, hookCfg HookConfig) error {
 
 	e.shell.Headerf("Running %s hook", hookName)
 
-	if !experiments.IsEnabled(ctx, experiments.PolyglotHooks) {
-		return e.runWrappedShellScriptHook(ctx, hookName, hookCfg)
-	}
-
 	hookType, err := hook.Type(hookCfg.Path)
 	if err != nil {
 		return fmt.Errorf("determining hook type for %q hook: %w", hookName, err)
@@ -408,7 +426,7 @@ func (e *Executor) runUnwrappedHook(ctx context.Context, hookName string, hookCf
 	environ.Set("BUILDKITE_HOOK_PATH", hookCfg.Path)
 	environ.Set("BUILDKITE_HOOK_SCOPE", hookCfg.Scope)
 
-	return e.shell.RunWithEnv(ctx, environ, hookCfg.Path)
+	return e.shell.Command(hookCfg.Path).Run(ctx, shell.WithExtraEnv(environ))
 }
 
 func logOpenedHookInfo(l shell.Logger, debug bool, hookName, path string) {
@@ -420,7 +438,7 @@ func logOpenedHookInfo(l shell.Logger, debug bool, hookName, path string) {
 		} else {
 			l.Errorf("The %s hook failed to run the %s process has the hook file open", hookName, procPath)
 		}
-	case utils.FileExists("/dev/fd"):
+	case osutil.FileExists("/dev/fd"):
 		isOpened, err := file.IsOpened(l, debug, path)
 		if err == nil {
 			if isOpened {
@@ -436,7 +454,7 @@ func logOpenedHookInfo(l shell.Logger, debug bool, hookName, path string) {
 	}
 }
 
-func logMissingHookInfo(l shell.Logger, wrapperPath string) {
+func logMissingHookInfo(l shell.Logger, hookName, wrapperPath string) {
 	// It's unlikely, but possible, that the script wrapper was spontaneously
 	// deleted or corrupted (it's usually in /tmp, which is fair game).
 	// A common setup error is to try to run a Bash hook in a container or other
@@ -445,7 +463,7 @@ func logMissingHookInfo(l shell.Logger, wrapperPath string) {
 	if err != nil {
 		// It's reasonable to assume the script wrapper was spontaneously
 		// deleted, or had something equally horrible happen to it.
-		l.Errorf("The %s hook failed to run - perhaps the wrapper script %q was spontaneously deleted", wrapperPath)
+		l.Errorf("The %s hook failed to run - perhaps the wrapper script %q was spontaneously deleted", hookName, wrapperPath)
 		return
 	}
 	interpreter := strings.TrimPrefix(shebang, "#!")
@@ -458,7 +476,7 @@ func logMissingHookInfo(l shell.Logger, wrapperPath string) {
 		// than ENOENT.
 		return
 	}
-	l.Errorf("The %s hook failed to run - perhaps the script interpreter %q is missing", interpreter)
+	l.Errorf("The %s hook failed to run - perhaps the script interpreter %q is missing", hookName, interpreter)
 }
 
 func (e *Executor) runWrappedShellScriptHook(ctx context.Context, hookName string, hookCfg HookConfig) error {
@@ -492,7 +510,7 @@ func (e *Executor) runWrappedShellScriptHook(ctx context.Context, hookName strin
 	const maxHookRetry = 30
 
 	// Run the wrapper script
-	if err := roko.NewRetrier(
+	err = roko.NewRetrier(
 		roko.WithStrategy(roko.Constant(100*time.Millisecond)),
 		roko.WithMaxAttempts(maxHookRetry),
 	).DoWithContext(ctx, func(r *roko.Retrier) error {
@@ -501,22 +519,29 @@ func (e *Executor) runWrappedShellScriptHook(ctx context.Context, hookName strin
 		// (which acquires open file descriptors of the parent process) and
 		// writing an executable (the script wrapper).
 		// See https://github.com/golang/go/issues/22315.
-		err := e.shell.RunScript(ctx, script.Path(), hookCfg.Env)
+		script, err := e.shell.Script(script.Path())
+		if err != nil {
+			r.Break()
+			return err
+		}
+		err = script.Run(ctx, shell.ShowPrompt(false), shell.WithExtraEnv(hookCfg.Env))
 		if errors.Is(err, syscall.ETXTBSY) {
 			return err
 		}
 		r.Break()
 		return err
-	}); err != nil {
-		exitCode := shell.GetExitCode(err)
+	})
+
+	if err != nil {
+		exitCode := shell.ExitCode(err)
 		e.shell.Env.Set("BUILDKITE_LAST_HOOK_EXIT_STATUS", strconv.Itoa(exitCode))
 
 		// If the hook exited with a non-zero exit code, then we should pass that back to the executor
 		// so it may inform the Buildkite API
 		if shell.IsExitError(err) {
 			return &shell.ExitError{
-				Code:    exitCode,
-				Message: fmt.Sprintf("The %s hook exited with status %d", hookName, exitCode),
+				Code: exitCode,
+				Err:  fmt.Errorf("The %s hook exited with status %d", hookName, exitCode),
 			}
 		}
 
@@ -531,7 +556,7 @@ func (e *Executor) runWrappedShellScriptHook(ctx context.Context, hookName strin
 			// program we tried to exec, even if the missing file/directory was
 			// actually the interpreter specified on the shebang line.
 			// Try to figure out which part is missing from the wrapper.
-			logMissingHookInfo(e.shell.Logger, script.Path())
+			logMissingHookInfo(e.shell.Logger, hookName, script.Path())
 		}
 
 		return err
@@ -629,29 +654,59 @@ func (e *Executor) applyEnvironmentChanges(changes hook.EnvChanges) {
 }
 
 func (e *Executor) hasGlobalHook(name string) bool {
-	_, err := e.globalHookPath(name)
-	return err == nil
+	_, err := hook.Find(e.HooksPath, name)
+	if err == nil {
+		return true
+	}
+	for _, additional := range e.AdditionalHooksPaths {
+		_, err := hook.Find(additional, name)
+		if err == nil {
+			return true
+		}
+	}
+	return false
 }
 
-// Returns the absolute path to a global hook, or os.ErrNotExist if none is found
-func (e *Executor) globalHookPath(name string) (string, error) {
-	return hook.Find(e.HooksPath, name)
+// find all matching paths for the specified hook
+func (e *Executor) getAllGlobalHookPaths(name string) ([]string, error) {
+	hooks := []string{}
+	p, err := hook.Find(e.HooksPath, name)
+	if err != nil {
+		if !os.IsNotExist(err) {
+			return []string{}, err
+		}
+	} else {
+		hooks = append(hooks, p)
+	}
+
+	for _, additional := range e.AdditionalHooksPaths {
+		p, err = hook.Find(additional, name)
+		// as this is an additional hook, don't fail if there's a problem here
+		if err == nil {
+			hooks = append(hooks, p)
+		}
+	}
+
+	return hooks, nil
 }
 
 // Executes a global hook if one exists
 func (e *Executor) executeGlobalHook(ctx context.Context, name string) error {
-	if !e.hasGlobalHook(name) {
+	allHooks, err := e.getAllGlobalHookPaths(name)
+	if err != nil {
 		return nil
 	}
-	p, err := e.globalHookPath(name)
-	if err != nil {
-		return err
+	for _, h := range allHooks {
+		err = e.executeHook(ctx, HookConfig{
+			Scope: "global",
+			Name:  name,
+			Path:  h,
+		})
+		if err != nil {
+			return err
+		}
 	}
-	return e.executeHook(ctx, HookConfig{
-		Scope: "global",
-		Name:  name,
-		Path:  p,
-	})
+	return nil
 }
 
 // Returns the absolute path to a local hook, or os.ErrNotExist if none is found
@@ -715,7 +770,7 @@ func dirForRepository(repository string) string {
 
 // Given a repository, it will add the host to the set of SSH known_hosts on the machine
 func addRepositoryHostToSSHKnownHosts(ctx context.Context, sh *shell.Shell, repository string) {
-	if utils.FileExists(repository) {
+	if osutil.FileExists(repository) {
 		return
 	}
 
@@ -758,7 +813,7 @@ func (e *Executor) setUp(ctx context.Context) error {
 	// The job runner sets BUILDKITE_IGNORED_ENV with any keys that were ignored
 	// or overwritten. This shows a warning to the user so they don't get confused
 	// when their environment changes don't seem to do anything
-	if ignored, exists := e.shell.Env.Get("BUILDKITE_IGNORED_ENV"); exists {
+	if ignored, _ := e.shell.Env.Get("BUILDKITE_IGNORED_ENV"); ignored != "" {
 		e.shell.Headerf("Detected protected environment variables")
 		e.shell.Commentf("Your pipeline environment has protected environment variables set. " +
 			"These can only be set via hooks, plugins or the agent configuration.")
@@ -910,7 +965,7 @@ func (e *Executor) CommandPhase(ctx context.Context) (hookErr error, commandErr
 	// error this will be zero. It's used to set the exit code later, so it's important
 	e.shell.Env.Set(
 		"BUILDKITE_COMMAND_EXIT_STATUS",
-		strconv.Itoa(shell.GetExitCode(commandErr)),
+		strconv.Itoa(shell.ExitCode(commandErr)),
 	)
 
 	// Exit early if there was no error
@@ -935,7 +990,7 @@ func (e *Executor) CommandPhase(ctx context.Context) (hookErr error, commandErr
 		// TODO: investigate phasing this out under a experiment
 		return nil, nil
 	case isExitError && !isExitSignaled:
-		e.shell.Errorf("The command exited with status %d", shell.GetExitCode(commandErr))
+		e.shell.Errorf("The command exited with status %d", shell.ExitCode(commandErr))
 		return nil, commandErr
 	default:
 		e.shell.Errorf("%s", commandErr)
@@ -965,7 +1020,7 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 
 	scriptFileName := strings.Replace(e.Command, "\n", "", -1)
 	pathToCommand, err := filepath.Abs(filepath.Join(e.shell.Getwd(), scriptFileName))
-	commandIsScript := err == nil && utils.FileExists(pathToCommand)
+	commandIsScript := err == nil && osutil.FileExists(pathToCommand)
 	span.AddAttributes(map[string]string{"hook.command": pathToCommand})
 
 	// If the command isn't a script, then it's something we need
@@ -985,19 +1040,19 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 
 	var cmdToExec string
 
-	// The shell gets parsed based on the operating system
-	shell, err := shellwords.Split(e.Shell)
+	// The interpreter gets parsed based on the operating system
+	interpreter, err := shellwords.Split(e.Shell)
 	if err != nil {
 		return fmt.Errorf("Failed to split shell (%q) into tokens: %w", e.Shell, err)
 	}
 
-	if len(shell) == 0 {
+	if len(interpreter) == 0 {
 		return fmt.Errorf("No shell set for job")
 	}
 
 	// Windows CMD.EXE is horrible and can't handle newline delimited commands. We write
 	// a batch script so that it works, but we don't like it
-	if strings.ToUpper(filepath.Base(shell[0])) == "CMD.EXE" {
+	if strings.ToUpper(filepath.Base(interpreter[0])) == "CMD.EXE" {
 		batchScript, err := e.writeBatchScript(e.Command)
 		if err != nil {
 			return err
@@ -1039,7 +1094,7 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 		// shouldn't be vulnerable to this!
 		if e.ExecutorConfig.CommandEval {
 			// Make script executable
-			if err = utils.ChmodExecutable(pathToCommand); err != nil {
+			if err = osutil.ChmodExecutable(pathToCommand); err != nil {
 				e.shell.Warningf("Error marking script %q as executable: %v", pathToCommand, err)
 				return err
 			}
@@ -1068,7 +1123,7 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 	}
 
 	var cmd []string
-	cmd = append(cmd, shell...)
+	cmd = append(cmd, interpreter...)
 	cmd = append(cmd, cmdToExec)
 
 	if e.Debug {
@@ -1077,7 +1132,7 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 		e.shell.Promptf("%s", cmdToExec)
 	}
 
-	err = e.shell.RunWithoutPrompt(ctx, cmd[0], cmd[1:]...)
+	err = e.shell.Command(cmd[0], cmd[1:]...).Run(ctx, shell.ShowPrompt(false))
 	return err
 }
 
@@ -1139,24 +1194,30 @@ func (e *Executor) writeBatchScript(cmd string) (string, error) {
 	return scriptFile.Name(), nil
 }
 
-// setupRedactors wraps shell output and logging in Redactors if any redaction
-// is necessary based on RedactedVars configuration and the existence of
-// matching environment vars. It will store the redactors in the Executor so
-// that they may be updated when the environment changes.
+// setupRedactors creates new stdout and [shell.Logger] to use for a new shell,
+// that write to stdout and stderr respectively, each via a [replacer.Replacer]
+// set up as a secret redactor. References to the redactors are retained in
+// e.redactors so they can be updated with new secrets.
 //
-// The returned method will remove the redactors from the Executor and Flush them.
-func (e *Executor) setupRedactors() {
-	varsToRedact, short, err := redact.Vars(e.ExecutorConfig.RedactedVars, e.shell.Env.DumpPairs())
+// Pictorally:
+//
+//	(returned io.Writer) == redactor 1 -> stdout
+//
+// and
+//
+//	(returned shell.Logger) -> redactor 2 -> stderr
+func (e *Executor) setupRedactors(log shell.Logger, environ *env.Environment, stdout, stderr io.Writer) (io.Writer, shell.Logger) {
+	varsToRedact, short, err := redact.Vars(e.ExecutorConfig.RedactedVars, environ.DumpPairs())
 	if err != nil {
-		e.shell.OptionalWarningf("bad-redacted-vars", "Couldn't match environment variable names against redacted-vars: %v", err)
+		log.OptionalWarningf("bad-redacted-vars", "Couldn't match environment variable names against redacted-vars: %v", err)
 	}
 	if len(short) > 0 {
 		slices.Sort(short)
-		e.shell.OptionalWarningf("short-redacted-vars", "Some variables have values below minimum length (%d bytes) and will not be redacted: %s", redact.LengthMin, strings.Join(short, ", "))
+		log.OptionalWarningf("short-redacted-vars", "Some variables have values below minimum length (%d bytes) and will not be redacted: %s", redact.LengthMin, strings.Join(short, ", "))
 	}
 
 	if e.Debug {
-		e.shell.Commentf("Enabling output redaction for values from environment variables matching: %v", e.ExecutorConfig.RedactedVars)
+		log.Commentf("Enabling output redaction for values from environment variables matching: %v", e.ExecutorConfig.RedactedVars)
 	}
 
 	needles := make([]string, 0, len(varsToRedact))
@@ -1164,46 +1225,16 @@ func (e *Executor) setupRedactors() {
 		needles = append(needles, pair.Value)
 	}
 
-	// If the shell Writer is already a Replacer, reset the values to redact.
-	if rdc, ok := e.shell.Writer.(*replacer.Replacer); ok {
-		// There may have been some redactees in the redactor already, so we
-		// propagate them to any new redactor.
-		needles = append(needles, rdc.Needles()...)
-		e.redactors.Append(rdc)
-	} else {
-		rdc := replacer.New(e.shell.Writer, needles, redact.Redact)
-		e.shell.Writer = rdc
-		e.redactors.Append(rdc)
-	}
+	stdoutRedactor := replacer.New(stdout, needles, redact.Redact)
+	e.redactors.Append(stdoutRedactor)
+	loggerRedactor := replacer.New(stderr, needles, redact.Redact)
+	e.redactors.Append(loggerRedactor)
 
-	// If the shell.Logger is already a redacted WriterLogger, reset the values to redact.
-	// (maybe there's a better way to do two levels of type assertion? ...
-	// shell.Logger may be a WriterLogger, and its Writer may be a Redactor)
-	var shellWriterLogger *shell.WriterLogger
-	var shellLoggerRedactor *replacer.Replacer
-	if logger, ok := e.shell.Logger.(*shell.WriterLogger); ok {
-		shellWriterLogger = logger
-		if redactor, ok := logger.Writer.(*replacer.Replacer); ok {
-			shellLoggerRedactor = redactor
-		}
-	}
-	if rdc := shellLoggerRedactor; rdc != nil {
-		// There may have been some redactees in the redactor already, so we
-		// propagate them to any new redactor.
-		needles = append(needles, rdc.Needles()...)
-		e.redactors.Append(rdc)
-	} else if shellWriterLogger != nil {
-		rdc := replacer.New(e.shell.Writer, needles, redact.Redact)
-		shellWriterLogger.Writer = rdc
-		e.redactors.Append(rdc)
-	}
-
-	e.redactors.Add(needles...)
+	logger := shell.NewWriterLogger(loggerRedactor, true, e.DisabledWarnings)
+	return stdoutRedactor, logger
 }
 
-func (e *Executor) kubernetesSetup(ctx context.Context, k8sAgentSocket *kubernetes.Client) error {
-	e.shell.Commentf("Using Kubernetes support")
-
+func (e *Executor) kubernetesSetup(ctx context.Context, environ *env.Environment, k8sAgentSocket *kubernetes.Client) error {
 	rtr := roko.NewRetrier(
 		roko.WithMaxAttempts(7),
 		roko.WithStrategy(roko.Exponential(2*time.Second, 0)),
@@ -1244,21 +1275,27 @@ func (e *Executor) kubernetesSetup(ctx context.Context, k8sAgentSocket *kubernet
 		switch n {
 		case "BUILDKITE_COMMAND", "BUILDKITE_ARTIFACT_PATHS", "BUILDKITE_PLUGINS":
 			continue
+
+		case "BUILDKITE_AGENT_ACCESS_TOKEN":
+			// Just in case someone has tried to fiddle with this, set it
+			// unconditionally (to be compatible with pre-v3.74.1 / PR 2851
+			// behavior).
+			environ.Set(n, v)
+			if err := os.Setenv(n, v); err != nil {
+				return err
+			}
+			continue
 		}
 		// Skip any that are already set.
-		if e.shell.Env.Exists(n) {
+		if environ.Exists(n) {
 			continue
 		}
 		// Set it!
-		e.shell.Env.Set(n, v)
+		environ.Set(n, v)
 		if err := os.Setenv(n, v); err != nil {
 			return err
 		}
 	}
-	// Attach the log stream to the k8s client
-	writer := io.MultiWriter(os.Stdout, k8sAgentSocket)
-	e.shell.Writer = writer
-	e.shell.Logger = shell.NewWriterLogger(writer, true, e.DisabledWarnings)
 
 	// Proceed when ready
 	if err := k8sAgentSocket.Await(ctx, kubernetes.RunStateStart); err != nil {
@@ -1268,10 +1305,14 @@ func (e *Executor) kubernetesSetup(ctx context.Context, k8sAgentSocket *kubernet
 	go func() {
 		// If the k8s client is interrupted because the "server" agent is
 		// stopped or unreachable, we should stop running the job.
+		err := k8sAgentSocket.Await(ctx, kubernetes.RunStateInterrupt)
 		// If the k8s client is interrupted because our own ctx was cancelled,
 		// then the job is already stopping, so there's no point logging an
 		// error.
-		if err := k8sAgentSocket.Await(ctx, kubernetes.RunStateInterrupt); err != nil && !errors.Is(err, context.Canceled) {
+		if errors.Is(err, context.Canceled) {
+			return
+		}
+		if err != nil {
 			e.shell.Errorf("Error waiting for client interrupt: %v", err)
 		}
 		e.Cancel()
